"""
AxApproval field type functions - before, after / insert, update, delete

If you want sequence - add children of previous step.
When user approves - he approves in all branches.
Field must contain list of pending actions.
When user added - AxMessage created in created branch with message about
    'You was invited to review and approve record'
When user approves - AxMessage created in all branches that user is set

Incase you want to email added person - you must do it in action python code
like so:
    ============================================================================

    initiator_email = ax.row.approval1.get('initiator')
    modifications = ax.row.approval1.get('modifications' , [])

    for mod in modifications:
        if mod['type'] == 'add':
            subject = f'⛏️ You was invited to approve - [{ax.tom_label}]'
            address = mod['reviewer']
            email_text = f"⛏️ You was invited to approve  - [{ax.tom_label}]"
            email_html = (
                f"⛏️ You was invited to approve Ax form - [{ax.tom_label}]"
                f"<br><a href='{ax.form_url}' >Open form [{ax.tom_label}]</a>"
            )
            await ax.email(
                to=address,
                text=email_text,
                html=email_html,
                subject=subject)

            ax.print(f'Email was sent to - {address} \n')
            ax.print(f'Current user email - {ax.user_email} \n')
            ax.print(f'Initiator email - {initiator_email} \n')

    ============================================================================

If you want a default approval path - you can create in in insert action like so

    field.value = {
        initiator: 'enf644@gmail.com',
        created: int(time.time()),
        branches: [{
            id: str(uuid.uuid4())
            reviewer: 'admin@someone.com',
            created: int(time.time()),
            status: 1,
            branches: [],
        }],
    }


Statuses:
    0 - waiting,
    1 - action needed
    3 - question added
    4 - approved
    5 - rejected

Value json structure:
    {
        initiator: 'enf644@gmail.com',
        created: 1574079223249,
        modified: 1574079223249,
        branches: [{
            id: '831d7d35-1b29-48bd-add0-e928e7c9e87f'
            reviewer: 'admin@someone.com',
            created: 1574079223249,
            modified: 1574079223249,
            status: 1,
            branches: [],
            resolution: 'I agree to this'
        }],
        awaitingActions: [
            {
                reviewer: 'admin@someone.com',
                created: 1574079223249
            }
        ],
        modifications: {
            {
                type: 'add',
                branchId: '831d7d35-1b29-48bd-add0-e928e7c9e87f',
                reviewer: 'boosov@gmail.com'
            },
            {
                type: 'status',
                branchId: '831d7d35-1b29-48bd-add0-e928e7c9e87f',
                reviewer: 'boosov@gmail.com'
                status: 3,
                resolution: 'I do not agree
            }
        }
    }

"""
import uuid
import time
from backend.model import AxMessage, AxMessageThread
import backend.misc as ax_misc
# import backend.emails as ax_emails


async def before_update(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Takes old value of field. Then using modifications modifies it.
        Approval statuses recieved from form are not used, only modifications.
        We walk over every branch of approval tree checking if action is needed

        Avalible actions are:
            add - invite new reviewer
            status - change branch status

        On every action the AxMessage is created for history.
    """
    del tobe_form, action
    current_user_email = current_user["email"]
    current_user_guid = current_user["user_id"]

    old_value = None
    if before_form.row_guid is None:
        old_value = field.value
    else:
        old_value = next(
            (fld.value for fld in before_form.fields if fld.guid == field.guid),
            None)
    modifications = field.value['modifications']

    async def create_message(thread_guid, text):
        thread = None
        if thread_guid and thread_guid != 'None':
            thread = db_session.query(AxMessageThread).filter(
                AxMessageThread.guid == uuid.UUID(str(thread_guid))
            ).first()
        if not thread:
            thread = AxMessageThread()
            thread.guid = uuid.UUID(str(thread_guid))
            db_session.add(thread)
            db_session.flush()

        new_message = AxMessage()
        new_message.author_guid = ax_misc.guid_or_none(current_user_guid)
        new_message.author_email = current_user_email
        new_message.text = text
        new_message.data_json = None
        new_message.thread_guid = thread.guid
        db_session.add(new_message)
        db_session.flush()

    async def check_if_action_needed(branch):
        """ Checks if modification list contains action for branch """
        now_date = int(time.time())
        if (not modifications or (
                current_user_email != branch.get('reviewer', None)
                and current_user_email != branch.get('initiator', None))):
            return branch
        for mod in modifications:
            if mod['branchId'] == branch['id']:
                already_presend = next(
                    (brch for brch in branch.get('branches')
                     if brch.get('reviewer') == mod.get('reviewer')),
                    None)

                text = None
                thread_guid = None
                new_branch_guid = str(uuid.uuid4())

                # Add new reviewer
                if mod['type'] == 'add' and not already_presend:

                    reviewer = mod['reviewer']
                    branch['branches'].append({
                        "id": new_branch_guid,
                        "reviewer": reviewer,
                        "created": now_date,
                        "status": 1,
                        "branches": [],
                    })
                    branch['modified'] = now_date
                    # TODO get this from locale!

                    # Add AxMessage
                    text = f'⛏️ [New reviewer was added]:\n {reviewer}'
                    thread_guid = branch['id'] or new_branch_guid
                    await create_message(thread_guid=thread_guid, text=text)

                    # If reviewer added to root branch - the branch_id is None
                    # and we cant add to thread. So we add this message to
                    # newly created branch
                    if not branch['id']:
                        thread_guid = new_branch_guid

                # Change status
                if mod['type'] == 'status':
                    resolution = mod['resolution']
                    status = int(mod['status'])
                    branch['status'] = status
                    branch['resolution'] = resolution
                    branch['modified'] = now_date
                    thread_guid = branch['id']

                    if status == 3:
                        text = f'⛏️ [Question was added]:\n {resolution}'
                    if status == 4:
                        text = f'⛏️ [Approval recieved]:\n {resolution}'
                    if status == 5:
                        text = f'⛏️ [Rejected]:\n {resolution}'

                    # Create AxMessage
                    await create_message(thread_guid=thread_guid, text=text)

        return branch

    async def check_node(branch):
        """ Check if modifications are needed for current branch and all
            chlidrens (recursive) """
        for idx, sub_brunch in enumerate(branch['branches']):
            branch['branches'][idx] = await check_node(branch=sub_brunch)

        branch = await check_if_action_needed(branch=branch)
        return branch

    # Walk over all nodes - check if action is needed
    field.value = await check_node(
        branch=old_value)

    #  In check_node we modified before_form. To pass new modifications
    # to action python and db, we need to rewrite it
    field.value["modifications"] = modifications

    field.needs_sql_update = True
    return field.value


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Create first value. Save initiator, then do before_update. """
    user_email = current_user["email"]
    now_date = int(time.time())
    new_branches = []

    if not field.value:
        return field.value

    for rev in field.value['branches']:
        # id will be replaced inside check_if_action_needed
        new_branches.append({
            "id": str(uuid.uuid4()),
            "reviewer": rev["reviewer"],
            "created": now_date,
            "status": 1,
            "branches": []
        })

    field.value = {
        "id": None,
        "initiator": user_email,
        "created": now_date,
        "branches": new_branches,
        "modifications": field.value["modifications"]
    }

    return await before_update(
        db_session=db_session,
        field=field,
        before_form=before_form,
        tobe_form=tobe_form,
        action=action,
        current_user=current_user)



"""AxAuthor
Before insert - write current user.
Before update - ignore change
after field is created - create dynamic role
before field is deleted - delete dynamic role


FORM
    get list of dynamic roles that user fits
    get allowed dict from cache
        normal - perm_{user_guid}_{field_guid}_{state_guid}
        dynamic - perm_{dynamic_role_guid}_{field_guid}_{state_guid}
                - action_perm_{dynamic_role_guid}_{action_guid}

GRID
    grid is same as roles on Start


"""
# import backend.model as ax_model
# from backend.model import AxRole


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Create default value of field.  """
    del db_session, before_form, tobe_form, action
    field.value = None
    if current_user:
        field.value = current_user["email"]
    return field.value


async def before_update(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Create default value of field.  """
    del db_session, before_form, tobe_form, action, current_user
    field.needs_sql_update = False
    return field.value


# async def after_field_create(db_session, field, before_form, tobe_form, action,
#                              current_user):
#     """ Create Author role for current form """
#     err = "AxAuthor -> after_create"
#     with ax_model.try_catch(db_session, err) as db_session:

#         # Check if form already contains Author role
#         for role in tobe_form.roles:
#             if role.field_tag == "AxAuthor":
#                 return False

#         ax_role = AxRole()
#         ax_role.name = "Author"  # TODO ADD python locale!
#         ax_role.form_guid = tobe_form.guid
#         ax_role.icon = "fas fa-user-tag"
#         ax_role.field_tag = "AxAuthor"
#         db_session.add(ax_role)
#         return True


# async def before_field_delete(db_session, field, before_form, tobe_form, action,
#                               current_user):
#     """ If field is removed - Delete Author role for current form """
#     # check if form contains more then one
#     author_fields_count = 0
#     for fld in tobe_form.fields:
#         if fld.field_type_tag == "AxAuthor":
#             author_fields_count += 1

#     if author_fields_count == 1:
#         db_session.query(AxRole).filter(
#             AxRole.form_guid == tobe_form.guid
#         ).filter(
#             AxRole.field_tag == "AxAuthor"
#         ).delete()


async def check_dynamic_role(db_session, ax_form, current_user):
    """ Checks if form values and current user fits current dynamic role
    If Author field == current_user_email -> return True """
    del db_session
    for field in ax_form.fields:
        if field.field_type_tag == "AxAuthor":
            if field.value == current_user['email']:
                return True
    return False



""" This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Workflow manipulation - create/update/delete - State, Action, Role,
Role2Action, Role2State, StatePermissions"""
import uuid
import graphene
# from loguru import logger
from backend.model import AxField, AxAction, AxState, \
    AxRole, AxRoleFieldPermission, AxState2Role, AxAction2Role, AxRole2Users
import backend.model as ax_model
from backend.schemas.types import State, Action, \
    State2Role, Action2Role, RoleFieldPermission, Role
from backend.auth import ax_admin_only
import backend.misc as ax_misc
import backend.auth as ax_auth


class CreateState(graphene.Mutation):
    """Create AxState

    Arguments:
        name (str): Name of created state. Usualy - default name from locale
        update (str): Default self-action name from locale. Like 'Update'
        form_guid (str): Guid of current AxForm
        x (float): X position of state in d3 workflow
        y (float): X position of state in d3 workflow

    Returns:
        state (AxState): Created state
        action (AxAction): Created update action
        permissions (List(AxRoleFieldPermission)): Default admin permissions
    """
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()
        update = graphene.String(required=False, default_value=None)
        form_guid = graphene.String()
        x = graphene.Float()
        y = graphene.Float()

    ok = graphene.Boolean()
    state = graphene.Field(State)
    action = graphene.Field(Action)
    permissions = graphene.List(RoleFieldPermission)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> CreateState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_form_guid = uuid.UUID(args.get('form_guid'))

            new_state = AxState()
            new_state.name = args.get('name')
            new_state.form_guid = ax_form_guid
            new_state.x = args.get('x')
            new_state.y = args.get('y')

            db_session.add(new_state)
            db_session.flush()

            # Create default self-action
            update_action = None
            if args.get('update'):
                update_action = AxAction()
                update_action.name = args.get('update')
                update_action.form_guid = ax_form_guid
                update_action.from_state_guid = new_state.guid
                update_action.to_state_guid = new_state.guid
                db_session.add(update_action)

            return CreateState(
                state=new_state,
                action=update_action,
                permissions=[],
                ok=True)


class UpdateState(graphene.Mutation):
    """Update AxState"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        x = graphene.Float()
        y = graphene.Float()

    ok = graphene.Boolean()
    state = graphene.Field(State)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> UpdateState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_state = db_session.query(AxState).filter(
                AxState.guid == uuid.UUID(args.get('guid'))
            ).first()
            if args.get('name'):
                ax_state.name = args.get('name')
                await ax_auth.write_dynamic_roles_cache(db_session)
            if args.get('x'):
                ax_state.x = args.get('x')
            if args.get('y'):
                ax_state.y = args.get('y')
            return UpdateState(state=ax_state, ok=True)


class DeleteState(graphene.Mutation):
    """Deletes AxState"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_state = db_session.query(AxState).filter(
                AxState.guid == uuid.UUID(guid)
            ).first()

            db_session.query(AxAction).filter(
                AxAction.from_state_guid == uuid.UUID(guid)
            ).delete()

            db_session.query(AxAction).filter(
                AxAction.to_state_guid == uuid.UUID(guid)
            ).delete()

            db_session.delete(ax_state)
            return DeleteState(deleted=guid, ok=True)


class CreateAction(graphene.Mutation):
    """Creates AxAction"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()
        from_state_guid = graphene.String()
        to_state_guid = graphene.String()

    ok = graphene.Boolean()
    action = graphene.Field(Action)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> CreateAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            new_action = AxAction()
            new_action.name = args.get('name')
            new_action.form_guid = args.get('form_guid')
            new_action.from_state_guid = args.get('from_state_guid')
            new_action.to_state_guid = args.get('to_state_guid')
            db_session.add(new_action)

            return CreateAction(action=new_action, ok=True)


class UpdateAction(graphene.Mutation):
    """Updates AxAction"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String(required=False, default_value=None)
        db_name = graphene.String(required=False, default_value=None)
        code = graphene.String(required=False, default_value=None)
        confirm_text = graphene.String(required=False, default_value=None)
        close_modal = graphene.Boolean(required=False, default_value=None)
        icon = graphene.String(required=False, default_value=None)
        radius = graphene.Float(required=False, default_value=None)
        job_options = graphene.String(required=False, default_value=None)

    ok = graphene.Boolean()
    action = graphene.Field(Action)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> UpdateAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_action = db_session.query(AxAction).filter(
                AxAction.guid == uuid.UUID(args.get('guid'))
            ).first()

            if args.get('name'):
                ax_action.name = args.get('name')
            if args.get('db_name'):
                ax_action.db_name = args.get('db_name')
            if args.get('code') is not None:
                if args.get('code') == '':
                    ax_action.code = None
                else:
                    ax_action.code = args.get('code')
            if args.get('confirm_text'):
                ax_action.confirm_text = args.get('confirm_text')
            if args.get('close_modal') is not None:
                ax_action.close_modal = args.get('close_modal')
            if args.get('icon'):
                ax_action.icon = args.get('icon')
            if args.get('radius') is not None:
                ax_action.radius = args.get('radius')

            return UpdateAction(action=ax_action, ok=True)


class DeleteAction(graphene.Mutation):
    """Deletes AxAction"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_action = db_session.query(AxAction).filter(
                AxAction.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_action)
            return DeleteAction(deleted=guid, ok=True)


def get_default_dynamic_role_code():
    """ Returns default code for dynamic role. """
    code = f"""ax.result = False
if(ax.row.some_field == ax.user_email):
    ax.result = True"""
    return code


class CreateRole(graphene.Mutation):
    """Creates AxRole"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()
        is_dynamic = graphene.Boolean()

    ok = graphene.Boolean()
    role = graphene.Field(Role)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> CreateRole"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_role = AxRole()
            ax_role.name = args.get('name')
            ax_role.form_guid = args.get('form_guid')
            ax_role.is_dynamic = args.get('is_dynamic')
            if ax_role.is_dynamic:
                ax_role.icon = 'fas fa-user-cog'
                ax_role.code = get_default_dynamic_role_code()
            db_session.add(ax_role)
            return CreateRole(role=ax_role, ok=True)


class UpdateRole(graphene.Mutation):
    """Updates AxRole"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String(required=False, default_value=None)
        icon = graphene.String(required=False, default_value=None)
        code = graphene.String(required=False, default_value=None)

    ok = graphene.Boolean()
    role = graphene.Field(Role)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> UpdateRole"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_role = db_session.query(AxRole).filter(
                AxRole.guid == uuid.UUID(args.get('guid'))
            ).first()
            if args.get('name'):
                ax_role.name = args.get('name')
            if args.get('icon'):
                ax_role.icon = args.get('icon')
            if args.get('code'):
                ax_role.code = args.get('code')

            return UpdateRole(role=ax_role, ok=True)


class DeleteRole(graphene.Mutation):
    """Deletes AxRole"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteRole"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')

            db_session.query(AxRole2Users).filter(
                AxRole2Users.role_guid == uuid.UUID(guid)
            ).delete()

            db_session.query(AxState2Role).filter(
                AxState2Role.role_guid == uuid.UUID(guid)
            ).delete()

            db_session.query(AxAction2Role).filter(
                AxAction2Role.role_guid == uuid.UUID(guid)
            ).delete()

            db_session.query(AxRoleFieldPermission).filter(
                AxRoleFieldPermission.role_guid == uuid.UUID(guid)
            ).delete()

            ax_role = db_session.query(AxRole).filter(
                AxRole.guid == uuid.UUID(guid)
            ).first()

            db_session.delete(ax_role)
            return DeleteRole(deleted=guid, ok=True)


class AddRoleToState(graphene.Mutation):
    """Create AxState2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        state_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()
    state2role = graphene.Field(State2Role)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> AddRoleToState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            role_guid = args.get('role_guid')
            state_guid = args.get('state_guid')
            existing_state2role = db_session.query(
                AxState2Role
            ).filter(
                AxState2Role.role_guid == uuid.UUID(role_guid)
            ).filter(
                AxState2Role.state_guid == uuid.UUID(state_guid)
            ).first()

            if existing_state2role:
                ok = True
                return AddRoleToState(state2role=existing_state2role, ok=ok)

            state2role = AxState2Role()
            state2role.state_guid = uuid.UUID(state_guid)
            state2role.role_guid = uuid.UUID(role_guid)

            db_session.add(state2role)
            return AddRoleToState(state2role=state2role, ok=True)


class DeleteRoleFromState(graphene.Mutation):
    """Deletes AxState2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        role_guid = graphene.String()
        state_guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()
    role_guid = graphene.String()
    state_guid = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteRoleFromState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            role_guid = args.get('role_guid')
            state_guid = args.get('state_guid')

            role2state = None
            if guid:
                role2state = db_session.query(AxState2Role).filter(
                    AxState2Role.guid == uuid.UUID(guid)
                ).first()
            else:
                role2state = db_session.query(AxState2Role).filter(
                    AxState2Role.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxState2Role.state_guid == uuid.UUID(state_guid)
                ).first()

            if role2state:
                the_role_guid = role2state.role_guid
                the_state_guid = role2state.state_guid

                db_session.delete(role2state)
                return DeleteRoleFromState(
                    deleted=guid,
                    role_guid=the_role_guid,
                    state_guid=the_state_guid,
                    ok=True)


class AddRoleToAction(graphene.Mutation):
    """Create AxAction2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        action_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()
    action2role = graphene.Field(Action2Role)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> AddRoleToAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            role_guid = args.get('role_guid')
            action_guid = args.get('action_guid')
            existing_action2role = db_session.query(
                AxAction2Role
            ).filter(
                AxAction2Role.role_guid == uuid.UUID(role_guid)
            ).filter(
                AxAction2Role.action_guid == uuid.UUID(action_guid)
            ).first()

            if existing_action2role:
                ok = True
                return AddRoleToAction(action2role=existing_action2role, ok=ok)

            action2role = AxAction2Role()
            action2role.action_guid = uuid.UUID(action_guid)
            action2role.role_guid = uuid.UUID(role_guid)

            db_session.add(action2role)
            return AddRoleToAction(action2role=action2role, ok=True)


class DeleteRoleFromAction(graphene.Mutation):
    """Deletes AxAction2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        action_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()
    role_guid = graphene.String()
    action_guid = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteRoleFromAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            action_guid = args.get('action_guid')
            role_guid = args.get('role_guid')

            role2action = None
            if guid:
                role2action = db_session.query(AxAction2Role).filter(
                    AxAction2Role.guid == uuid.UUID(guid)
                ).first()
            else:
                role2action = db_session.query(AxAction2Role).filter(
                    AxAction2Role.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxAction2Role.action_guid == uuid.UUID(action_guid)
                ).first()

            deleted = role2action.guid
            db_session.delete(role2action)
            return DeleteRoleFromAction(
                deleted=deleted,
                ok=True,
                role_guid=role_guid,
                action_guid=action_guid)


class SetStatePermission(graphene.Mutation):
    """Creates AxRoleFieldPermission or multiple if tab field is presented"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        state_guid = graphene.String()
        role_guid = graphene.String()
        field_guid = graphene.String(required=False, default_value=None)
        read = graphene.Boolean()
        edit = graphene.Boolean()

    ok = graphene.Boolean()
    permissions = graphene.List(RoleFieldPermission)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> SetStatePermission"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            state_guid = args.get('state_guid')
            role_guid = args.get('role_guid')
            field_guid = args.get('field_guid')
            read = args.get('read')
            edit = args.get('edit')

            ax_field_guid = None
            ax_field = None
            if field_guid:
                ax_field = db_session.query(AxField).filter(
                    AxField.guid == uuid.UUID(field_guid)).first()
                ax_field_guid = ax_field.guid

            # If field_guid is null, then access is set for whole form
            # We must delete all fields permissions
            if not ax_field:
                db_session.query(
                    AxRoleFieldPermission
                ).filter(
                    AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
                ).filter(
                    AxRoleFieldPermission.role_guid == uuid.UUID(role_guid)
                ).delete()
            else:
                # if field or tab clicked - we must delete
                # permission fo all fields
                db_session.query(
                    AxRoleFieldPermission
                ).filter(
                    AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
                ).filter(
                    AxRoleFieldPermission.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxRoleFieldPermission.field_guid.is_(None)
                ).delete()

            # If field is tab - web must delete all permission for fields
            # of this tab
            if ax_field and ax_field.is_tab:
                tab_fields = db_session.query(AxField).filter(
                    AxField.parent == ax_field_guid
                ).all()
                tab_fields_guids = [fld.guid for fld in tab_fields]

                tab_perms = db_session.query(
                    AxRoleFieldPermission
                ).filter(
                    AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
                ).filter(
                    AxRoleFieldPermission.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxRoleFieldPermission.field_guid.in_(tab_fields_guids)
                ).all()

                for perm in tab_perms:
                    db_session.delete(perm)

            # If field is not tab - web must delete tab permission
            if ax_field and not ax_field.is_tab:
                db_session.query(
                    AxRoleFieldPermission
                ).filter(
                    AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
                ).filter(
                    AxRoleFieldPermission.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxRoleFieldPermission.field_guid == ax_field.parent
                ).delete()

            ax_perm = db_session.query(
                AxRoleFieldPermission
            ).filter(
                AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
            ).filter(
                AxRoleFieldPermission.role_guid == uuid.UUID(role_guid)
            ).filter(
                AxRoleFieldPermission.field_guid == ax_field_guid
            ).first()

            if ax_perm is None:
                ax_perm = AxRoleFieldPermission()
                ax_perm.form_guid = uuid.UUID(form_guid)
                ax_perm.role_guid = uuid.UUID(role_guid)
                ax_perm.field_guid = ax_field_guid
                ax_perm.state_guid = uuid.UUID(state_guid)
                ax_perm.read = read
                ax_perm.edit = edit
                db_session.add(ax_perm)
            else:
                ax_perm.read = read
                ax_perm.edit = edit

            db_session.flush()

            return_permissions = db_session.query(
                AxRoleFieldPermission
            ).filter(
                AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
            ).all()

            return SetStatePermission(permissions=return_permissions, ok=True)


class WorkflowQuery(graphene.ObjectType):
    """Workflow query"""
    ax_role = graphene.Field(
        Role,
        guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    @ax_admin_only
    async def resolve_ax_role(self, info, guid, update_time=None):
        """Get AxRole with code"""
        del update_time
        err = 'Error in GQL query - resolve_ax_role.'
        with ax_model.try_catch(info.context['session'], err, no_commit=True):
            query = Role.get_query(info)  # SQLAlchemy query
            role_guid = ax_misc.guid_or_none(guid)
            ret_role = query.filter(
                AxRole.guid == role_guid
            ).first()
            return ret_role


class WorkflowMutations(graphene.ObjectType):
    """Combine all mutations"""
    create_state = CreateState.Field()
    update_state = UpdateState.Field()
    delete_state = DeleteState.Field()
    create_action = CreateAction.Field()
    update_action = UpdateAction.Field()
    delete_action = DeleteAction.Field()
    create_role = CreateRole.Field()
    update_role = UpdateRole.Field()
    delete_role = DeleteRole.Field()
    add_role_to_state = AddRoleToState.Field()
    delete_role_from_state = DeleteRoleFromState.Field()
    add_role_to_action = AddRoleToAction.Field()
    delete_role_from_action = DeleteRoleFromAction.Field()
    set_state_permission = SetStatePermission.Field()



""" Executes python code. Used in some field types, in dynamic roles """
# import os
import sys
# from loguru import logger
from dotmap import DotMap

import backend.dialects as ax_dialects
import backend.misc as ax_misc

this = sys.modules[__name__]


async def aexec(code, localz, **kwargs):
    """ This function wraps python code from AxAction with async function
    and runs it """
    # Restore globals later
    args = ", ".join(list(kwargs.keys()))
    code_from_action = code.replace("\n", "\n    ")
    async_code = (f"async def func({args}):"
                  f"\n    {code_from_action}"
                  f"\n    return ax")
    exec(async_code, {}, localz)    # pylint: disable=exec-used
    # Don't expect it to return from the coro.
    result = await localz["func"](**kwargs)
    return result


async def execute_field_code(code, form, arguments=None, current_user=None):
    """ Used to execute AxField backend code. see /backend/fields for info """
    localz = dict()
    host = await ax_misc.get_ax_host()
    ax = DotMap()  # javascript style dicts item['guid'] == item.guid
    ax.row.guid = form.row_guid
    ax.form = form
    ax.host = host
    ax.sql = ax_dialects.dialect.custom_query
    ax.user_email = None
    ax.user_guid = None
    if current_user:
        ax.user_email = current_user.get("email", None)
        ax.user_guid = current_user.get("user_id", None)
    for field in form.db_fields:
        ax.row[field.db_name] = field.value
    if arguments:
        for key, value in arguments.items():
            ax[key] = value
    localz['ax'] = ax

    try:
        await aexec(code=str(code), localz=localz, ax=ax)
        ret_ax = localz['ax']
        return ret_ax
    except SyntaxError as err:
        return err
    except Exception as err:    # pylint: disable=broad-except
        return err



"""AxNum field type functions - before, after / insert, update, delete"""
import json
from backend.model import AxMetric
import backend.exec as ax_exec


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """
    Executes python code from AxField.private_options

    Returns:
        Object: Returns updated value of current field"""
    del before_form, action

    if not field.private_options_json or field.private_options_json == '{}':
        return None

    # get current counter from AxMetric for current key
    # if there is none -> create one

    options = json.loads(field.private_options_json)
    code = options['algorithm']
    key = options['counterKey']

    if not key:
        return None

    current_counter = db_session.query(AxMetric).filter(
        AxMetric.key == key
    ).first()

    current_value = None
    if current_counter:
        current_value = current_counter.value
    else:
        current_counter = AxMetric()
        current_counter.key = key
        current_counter.value = None
        db_session.add(current_counter)

    arguments = {
        "counter": current_value
    }

    ax = await ax_exec.execute_field_code(
        code=code,
        form=tobe_form,
        arguments=arguments,
        current_user=current_user)

    field.needs_sql_update = True
    field.value = ax.value
    current_counter.value = ax.counter

    return field.value



"""All routes of ax-core and ax-admin"""

import os
import sys
import asyncio
import uuid
import json
from sanic import response
from loguru import logger
from graphql.execution.executors.asyncio import AsyncioExecutor
from graphql_ws.websockets_lib import WsLibSubscriptionServer
from sanic_jwt.decorators import inject_user
import markdown2

from backend.graphqlview import GraphQLView
import backend.cache as ax_cache
import backend.schema as ax_schema
import backend.misc as ax_misc
from backend.tus import tus_bp
import backend.model as ax_model
from backend.model import AxForm, AxField
import backend.schemas.form_schema as form_schema
import backend.dialects as ax_dialects
import backend.auth as ax_auth
import backend.migration as ax_migration
from backend.auth import ax_protected


this = sys.modules[__name__]

loop = asyncio.new_event_loop()
actions_loop = asyncio.new_event_loop()
app = None
graphql_view = None


def init_graphql_view():  # pylint: disable=unused-variable
    """Initiate graphql"""
    this.graphql_view = GraphQLView.as_view(
        schema=ax_schema.schema,
        graphiql=False,
        enable_async=True,
        executor=AsyncioExecutor(loop=this.loop)
    )
    this.app.add_route(this.graphql_view, '/api/graphql')


def init_routes(sanic_app, pages_path=None, ssl_enabled=False):  # pylint: disable=unused-variable
    """Innitiate all Ax routes"""
    del ssl_enabled
    try:
        sanic_app.static('/uploads', str(ax_misc.path('uploads')))
        sanic_app.static('/static', str(ax_misc.path('dist/ax/static')))
        sanic_app.static('/stats', str(ax_misc.path('dist/ax/stats.html')))
        sanic_app.static(
            '/test_webpack', str(ax_misc.path('dist/ax/test.html')))

        sanic_app.static(
            '/editor.worker.js', str(ax_misc.path('dist/ax/editor.worker.js')))
        sanic_app.static(
            '/html.worker.js', str(ax_misc.path('dist/ax/html.worker.js')))
        sanic_app.static(
            'json.worker.js', str(ax_misc.path('dist/ax/json.worker.js')))

        # Pages routes {
        pages_dist_path = str(ax_misc.path('dist/pages'))
        if pages_path:
            pages_dist_path = pages_path
        index_path = str(os.path.join(pages_dist_path, "index.html"))

        sanic_app.static(
            '/pages/static', str(ax_misc.path('dist/pages/static')))
        # TODO WTF this guid? maybe must be part of build?

        pre_cache_file_name = None
        pages_dist_dir = ax_misc.path('dist/pages')
        file_list = os.listdir(pages_dist_dir)
        for name in file_list:
            if "precache-manifest." in name:
                pre_cache_file_name = name

        sanic_app.static(
            f'/pages/{pre_cache_file_name}',
            str(ax_misc.path(f'dist/pages/{pre_cache_file_name}')))     # pylint: disable=line-too-long
        sanic_app.static(
            '/pages/service-worker.js',
            str(ax_misc.path('dist/pages/service-worker.js')))
        sanic_app.static(
            '/pages/manifest.json',
            str(ax_misc.path('dist/pages/manifest.json')))
        sanic_app.static(
            '/pages/robots.txt',
            str(ax_misc.path('dist/pages/robots.txt')))

        sanic_app.static('/pages', index_path)

        @sanic_app.route('/pages/<path:path>')
        def pages_index(request, path=None):  # pylint: disable=unused-variable
            """  """
            del request, path
            return response.html(open(index_path).read())

        @sanic_app.route('/signin')
        def pages_index1(request, path=None):  # pylint: disable=unused-variable
            """  """
            del request, path
            return response.html(open(index_path).read())

        # Pages routes }

        # Add tus upload blueprint
        sanic_app.blueprint(tus_bp)
        # Add web-socket subscription server
        subscription_server = WsLibSubscriptionServer(ax_schema.schema)

        @sanic_app.route('api/signout', methods=['GET'])
        @inject_user()
        @ax_protected()
        async def signout(request, user=None):   # pylint: disable=unused-variable
            """ Delete all auth cookies and redirect to signin """

            user_guid = user.get('user_id', None) if user else None
            if user_guid:
                key = f'refresh_token_{user_guid}'
                await ax_cache.cache.delete(key)

            to_url = '/signin'
            if request.args.get('to_admin', False):
                to_url = '/admin/signin'

            resp = response.redirect(to_url)
            del resp.cookies['ax_auth']
            del resp.cookies['access_token']
            del resp.cookies['refresh_token']
            return resp

        @sanic_app.route(
            '/api/file/<form_guid>/<row_guid>/<field_guid>/<file_name>',
            methods=['GET'])
        @inject_user()
        @ax_protected()
        async def db_file_viewer(   # pylint: disable=unused-variable
                request, form_guid, row_guid, field_guid, file_name, user):  # pylint: disable=unused-variable
            """ Used to display files that are stored in database.
                Used in fields like AxImageCropDb"""
            current_user = user
            del request, form_guid, file_name
            with ax_model.scoped_session("routes.db_file_viewer") as db_session:
                safe_row_guid = str(uuid.UUID(str(row_guid)))
                ax_field = db_session.query(AxField).filter(
                    AxField.guid == uuid.UUID(field_guid)
                ).first()

                # first we select only guid and axState to know, if user have
                # access
                row_result = await ax_dialects.dialect.select_one(
                    db_session=db_session,
                    form=ax_field.form,
                    fields_list=[],
                    row_guid=safe_row_guid)

                state_name = row_result[0]['axState']
                state_guid = await ax_auth.get_state_guid(
                    ax_form=ax_field.form,
                    state_name=state_name)

                user_guid = current_user.get(
                    'user_id', None) if current_user else None
                user_is_admin = current_user.get(
                    'is_admin', False) if current_user else False
                allowed_field_dict = await ax_auth.get_allowed_fields_dict(
                    ax_form=ax_field.form,
                    user_guid=user_guid,
                    state_guid=state_guid)

                field_guid = str(ax_field.guid)
                if not user_is_admin:
                    if (field_guid not in allowed_field_dict or
                            allowed_field_dict[field_guid] == 0 or
                            allowed_field_dict[field_guid] is None):
                        email = current_user.get('email', None)
                        msg = (
                            f'Error in db_file_viewer. ',
                            f'not allowed for user [{email}]'
                        )
                        logger.error(msg)
                        return response.text("", status=403)

                field_value = await ax_dialects.dialect.select_field(
                    db_session=db_session,
                    form_db_name=ax_field.form.db_name,
                    field_db_name=ax_field.db_name,
                    row_guid=safe_row_guid)

                return response.raw(
                    field_value, content_type='application/octet-stream')

        @sanic_app.route(
            '/api/file/<form_guid>/<row_guid>/<field_guid>/<file_guid>/<file_name>',    # pylint: disable=line-too-long
            methods=['GET'])
        @inject_user()
        @ax_protected()
        async def file_viewer(  # pylint: disable=unused-variable
                request, form_guid, row_guid, field_guid, file_guid, file_name,\
                user=None):
            """ Used to display files uploaded and stored on disk.
                Displays temp files too. Used in all fields with upload"""
            del request
            current_user = user
            with ax_model.scoped_session("routes -> file_viewer") as db_session:
                # if row_guid is null -> display from /tmp without permissions
                if not row_guid or row_guid == 'null':
                    tmp_dir = os.path.join(ax_misc.tmp_root_dir, file_guid)
                    file_name = os.listdir(tmp_dir)[0]
                    temp_path = os.path.join(tmp_dir, file_name)
                    return await response.file(temp_path)

                # get AxForm with row values
                ax_form = db_session.query(AxForm).filter(
                    AxForm.guid == uuid.UUID(form_guid)
                ).first()
                ax_form = await form_schema.set_form_values(
                    db_session=db_session,
                    ax_form=ax_form,
                    row_guid=row_guid,
                    current_user=current_user)

                # Get values from row, field
                field_values = None
                for field in ax_form.fields:
                    if field.guid == uuid.UUID(field_guid):
                        if field.value:
                            field_values = json.loads(field.value)

                            if type(field_values) is str:
                                try:
                                    field_values = json.loads(field_values)
                                except:
                                    pass

                # Find requested file in value
                the_file = None
                for file in field_values:
                    if file['guid'] == file_guid:
                        the_file = file

                if not the_file:
                    return response.text("", status=404)

                state_guid = await ax_auth.get_state_guid(
                    ax_form=ax_form,
                    state_name=ax_form.current_state_name)

                user_guid = current_user.get(
                    'user_id', None) if current_user else None
                user_is_admin = current_user.get(
                    'is_admin', False) if current_user else False
                allowed_field_dict = await ax_auth.get_allowed_fields_dict(
                    ax_form=ax_form,
                    user_guid=user_guid,
                    state_guid=state_guid)

                if not user_is_admin:
                    if (field_guid not in allowed_field_dict or
                            allowed_field_dict[field_guid] == 0 or
                            allowed_field_dict[field_guid] is None):
                        email = current_user['email']
                        msg = (
                            f'Error in file_viewer. ',
                            f'not allowed for user [{email}]'
                        )
                        logger.error(msg)
                        return response.text("", status=403)

                # if file exists -> return file
                row_guid_str = str(uuid.UUID(row_guid))
                file_path = os.path.join(
                    ax_misc.uploads_root_dir,
                    'form_row_field_file',
                    form_guid,
                    row_guid_str,
                    field_guid,
                    the_file['guid'],
                    the_file['name'])
                if not os.path.lexists(file_path):
                    return response.text("", status=404)
                return await response.file(file_path)

        @sanic_app.route('/admin/<path:path>')
        def index(request, path=None):  # pylint: disable=unused-variable
            """ This is MAIN ROUTE. (except other routes listed in this module).
                All requests are directed to Vue single page app. After that Vue
                handles routing."""
            del request, path
            absolute_path = ax_misc.path('dist/ax/index.html')
            return response.html(open(absolute_path).read())

        @sanic_app.route('/form/<path:path>')
        def index1(request, path=None):  # pylint: disable=unused-variable
            """ Copy of index. Sanic bug - https://github.com/huge-success/sanic/pull/1779"""
            del request, path
            absolute_path = ax_misc.path('dist/ax/index.html')
            return response.html(open(absolute_path).read())

        @sanic_app.route('/grid/<path:path>')
        def index2(request, path=None):  # pylint: disable=unused-variable
            """ Copy of index. Sanic bug - https://github.com/huge-success/sanic/pull/1779"""
            del request, path
            absolute_path = ax_misc.path('dist/ax/index.html')
            return response.html(open(absolute_path).read())

        @sanic_app.route('/admin/signin')
        def index3(request, path=None):  # pylint: disable=unused-variable
            """ Copy of index. Sanic bug - https://github.com/huge-success/sanic/pull/1779"""
            del request, path
            absolute_path = ax_misc.path('dist/ax/index.html')
            return response.html(open(absolute_path).read())

        @sanic_app.route('/')
        def handle_request(request):  # pylint: disable=unused-variable
            del request
            return response.redirect('/pages')

        @sanic_app.route('/api/draw_ax')
        async def draw_ax(request):  # pylint: disable=unused-variable
            """ Outputs bundle.js. Used when Ax web-components
                are inputed somewhere. Users can use this url for <script> tag
                """
            del request
            absolute_path = ax_misc.path('dist/ax/static/js/ax-bundle.js')
            return await response.file(
                absolute_path,
                headers={
                    'Content-Type': 'application/javascript; charset=utf-8'
                }
            )

        @sanic_app.route('/draw_ax')
        async def draw_ax1(request):  # pylint: disable=unused-variable
            """ Outputs bundle.js. Used when Ax web-components
                are inputed somewhere. Users can use this url for <script> tag
                """
            del request
            absolute_path = ax_misc.path('dist/ax/static/js/ax-bundle.js')
            return await response.file(
                absolute_path,
                headers={
                    'Content-Type': 'application/javascript; charset=utf-8'
                }
            )

        @sanic_app.websocket('/api/subscriptions', subprotocols=['graphql-ws'])
        async def subscriptions(request, web_socket):  # pylint: disable=unused-variable
            """Web socket route for graphql subscriptions"""
            del request
            try:
                # TODO: Why socket error exception occurs without internet
                await subscription_server.handle(web_socket)
                return web_socket
            except asyncio.CancelledError:
                pass
                # logger.exception('Socket error')

        @sanic_app.route('/api/ping', methods=['GET', 'HEAD'])
        async def ping(request):  # pylint: disable=unused-variable
            """ Ping function checks if Ax is running. Used with monit """
            del request
            from backend.schema import schema
            result = schema.execute("query { ping }")
            test_str = json.dumps(result.data, sort_keys=True, indent=4)
            return response.text(test_str)

        @sanic_app.route('/api/blog_rss')
        async def blog_rss(request):  # pylint: disable=unused-variable
            """ Fetches medium blog rss """
            del request
            feed = await ax_misc.fetch('https://medium.com/feed/@enf644')   # pylint: disable=line-too-long
            return response.text(feed)

        @sanic_app.route('/api/stackoverflow_rss')
        async def stackoverflow_rss(request):  # pylint: disable=unused-variable
            """ Fetches stackoverflow tag rss """
            del request
            feed = await ax_misc.fetch('https://stackexchange.com/feeds/tagsets/381786/ax-workflow?sort=active')   # pylint: disable=line-too-long
            return response.text(feed)

        @sanic_app.route('/api/home_welcome')
        async def home_welcome(request):  # pylint: disable=unused-variable
            """ Fetches welcome_free.md from ax-info """
            del request
            the_url = 'https://raw.githubusercontent.com/enf644/ax-info/master/welcome_free.md'
            if ax_auth.lise_is_active():
                the_url = 'https://raw.githubusercontent.com/enf644/ax-info/master/welcome.md'
            feed = await ax_misc.fetch(the_url)   # pylint: disable=line-too-long
            html = markdown2.markdown(feed)
            return response.text(html)

        @sanic_app.route('/api/marketplace_featured')
        async def marketplace_featured(request):  # pylint: disable=unused-variable
            """ Fetches featured apps json """
            del request
            feed = await ax_misc.fetch('https://raw.githubusercontent.com/enf644/ax-info/master/featured_apps.json')   # pylint: disable=line-too-long
            return response.text(feed)

        @sanic_app.route('/api/marketplace_all')
        async def marketplace_apps(request):  # pylint: disable=unused-variable
            """ Fetches all apps json """
            del request
            feed = await ax_misc.fetch('https://raw.githubusercontent.com/enf644/ax-info/master/apps.json')   # pylint: disable=line-too-long
            return response.text(feed)

        @sanic_app.route('/api/test')
        async def test(request):  # pylint: disable=unused-variable
            """Test function"""
            del request

            ret_str = await ax_migration.send_stats()

            # data = {
            #     "host": "hostHost",
            #     "usersNum": 10
            # }
            # value_str = json.dumps(data).replace('"', '\\"')
            # query_str = (
            #     "    mutation{"
            #     "        doAction("
            #     "            formDbName: \"AxUsageStats\""
            #     "            actionDbName: \"newStats\""
            #     f"            values: \"{value_str}\""
            #     "        ) {"
            #     "            ok"
            #     "        }"
            #     "    }"
            # )

            # json_data = {'query': query_str}
            # ret_str = await ax_misc.post_json(
            #     'http://127.0.0.1:8080/api/graphql', json_data)

            # ret_str = ax_model.engine.pool.status()
            # this.test_schema = 'IT WORKS'
            # ax_pubsub.publisher.publish(
            #     aiopubsub.Key('dummy_test'), this.test_schema)
            return response.text(ret_str)

        @sanic_app.route('/api/set')
        async def cache_set(request):  # pylint: disable=unused-variable
            """Cache Test function"""
            del request
            obj = ['one', 'two', 'three']
            await ax_cache.cache.set('user_list', obj)
            return response.text('Cache SET' + str(obj))

        @sanic_app.route('/api/get')
        async def cache_get(request):  # pylint: disable=unused-variable
            """Cache Test function"""
            del request
            obj = await ax_cache.cache.get('user_list')
            ret_str = 'READ cache == ' + \
                str(obj[0].username + ' - ' + os.environ['AX_VERSION'])
            return response.text(ret_str)

    except Exception:
        logger.exception('Error initiating routes.')
        raise



"""AxPaymentStripe

For AxPaymentStripe to work you need to create intent in AxAction python
code like so  (stripe1)-

    ax.row.stripe['intent'] = ax.stripe.PaymentIntent.create(
        amount=500,
        currency='usd',
    )

WARNING! Amount must be multipleed by 100! amount=500 meens 5$

4242424242424242	Succeeds and immediately processes the payment.
4000002500003155	Requires authentication.
                    Stripe will trigger a modal asking for
                    the customer to authenticate.
4000000000009995	Always fails with a decline code of insufficient_funds.

"""
import os


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Create default value of field.  """
    del db_session, before_form, tobe_form, action, current_user
    field.value = {
        "pubKey": os.getenv('STRIPE_PUBLISHABLE_KEY'),
        "intent": None
    }
    field.force_sql_update = True
    return field.value



"""Miscellaneous functions
"""

import os
import sys
import re
import uuid
import zipfile
import shutil
from pathlib import Path
from datetime import datetime
import asyncio
import graphene
# import graphene_sqlalchemy
import pytz
import yaml
from loguru import logger
# from dotmap import DotMap
import aiohttp

this = sys.modules[__name__]
root_path = None
timezone = None
tmp_root_dir = None
uploads_root_dir = None


def init_misc(timezone_name, tmp_absolute_path, uploads_absolute_path) -> None:
    """Inititate time_zone, root_path"""

    try:
        this.uploads_root_dir = path('uploads')
        if uploads_absolute_path is not None:
            this.uploads_root_dir = str(Path(uploads_absolute_path))
        if not server_is_app_engine():
            if not os.path.exists(this.uploads_root_dir):
                os.makedirs(this.uploads_root_dir)

        if not os.access(this.uploads_root_dir, os.W_OK):
            raise PermissionError(
                f'Cant write to directory - {this.uploads_root_dir}')
    except Exception:
        logger.exception(
            'Uploads folder does not exist and Ax cant create it')
        raise

    try:
        this.tmp_root_dir = path('tmp')
        if tmp_absolute_path is not None:
            this.tmp_root_dir = str(Path(tmp_absolute_path))
        if not os.path.exists(this.tmp_root_dir):
            if not server_is_app_engine():
                os.makedirs(this.tmp_root_dir)
    except Exception:
        logger.exception(
            'Tmp folder does not exist and Ax cant create it')
        raise

    if timezone_name not in pytz.common_timezones:
        logger.error(
            'Timezone {tz} not valid. Falling back to UTC.', tz=timezone_name)
        this.timezone = 'UTC'

    this.timezone = pytz.timezone(timezone_name)


def date(_date: datetime = datetime.now()) -> datetime:
    """Localise python datetime"""
    localized_date = this.timezone.localize(_date, is_dst=None)
    return localized_date


def path(_path: str = '') -> str:
    """Get correct absolute path.

    Agrs:
        _path (str, optional): Defaults to ''. Path relative to module root.

    Returns:
        str: Absolute path to specified file or directory
    Examples:
        Access module files like path('backend/schemas/user_schema.py')
    """
    if this.root_path is None:
        this.root_path = Path(__file__).resolve().parent.parent
        # logger.debug('File path = {root}', root=Path(__file__))
    return root_path / _path


def server_is_app_engine() -> bool:
    """Is Sanic is running on Google App Engine Standard"""
    return os.getenv("GAE_VERSION") or None


def load_configuration() -> None:
    """Function that loads ax application configuration from app.yaml

    If application is running on Google App Engine, all params from
    env_variables section of app.yaml will be loaded as enviroment variables.
    If application is running outside of App Engine then we must load
    variables ourselves.

    Args:
        app_yaml_location (str): Path to app.yaml location.

    Returns:
        bool: The return value. True for success.
    Examples:
        Access configuration variables like os.environ['AX_VERSION']

    Raises:
        FileNotFoundError: Cant find app.yaml
        ValueError: Cant parse app.yaml
        LookupError: No env_variables section in app.yaml
    """
    app_yaml = path('app.yaml')
    app_template_yaml = path('app.template.yaml')

    if not os.path.isfile(app_yaml):
        # os.rename(app_template_yaml, app_yaml)
        shutil.copyfile(app_template_yaml, app_yaml)

    if not os.path.isfile(app_yaml):
        raise FileNotFoundError('Configuration failed, app.yaml not found')

    if not server_is_app_engine():
        with open(app_yaml, 'r') as stream:
            try:
                yaml_vars = yaml.safe_load(stream)
                if 'env_variables' not in yaml_vars:
                    err = 'Configuration failed, no env_variables in app.yaml'
                    logger.error(err)
                    raise LookupError(err)
                for key, value in yaml_vars['env_variables'].items():
                    if value:
                        os.environ[key] = str(value)
            except yaml.YAMLError as exc:
                err = 'Configuration failed, cant parse yaml - ' + exc
                logger.error(err)
                raise ValueError(err)


def convert_column_to_string(type, column, registry=None):  # pylint: disable=redefined-builtin
    """This functions is needed to convert UUID column to String column
    for SQL Alchemy"""
    del type, column, registry
    return graphene.String


def string_is_guid(string_guid):
    """Checks if string is valid guid can be
     302a4299736e4ef384fca9f400e84b24 or 302a4299-736e-4ef3-84fc-a9f400e84b24

    Returns:
        [str]: version of uuid or None
    """
    try:
        return uuid.UUID(str(string_guid)).version
    except ValueError:
        return None


def guid_or_none(string_guid):
    """ returns uuid or None """
    if string_is_guid(string_guid):
        return uuid.UUID(str(string_guid))
    return None


async def zip_folder(output_filename, source_dir):
    """ Archive folder """
    relroot = source_dir
    with zipfile.ZipFile(output_filename, "w", zipfile.ZIP_DEFLATED) as z_file:
        for root, dirs, files in os.walk(source_dir):
            del dirs
            # add directory (needed for empty dirs)
            # zip.write(root, os.path.relpath(root, relroot))
            for file in files:
                filename = os.path.join(root, file)
                if os.path.isfile(filename):  # regular files only
                    arcname = os.path.join(
                        os.path.relpath(root, relroot), file)
                    z_file.write(filename, arcname)


async def get_tom_label(form):
    """ Gets AxForm with field values and constructs tom_label by replacing
        {{tags}} with field values """
    tom_label_modified = form.tom_label.replace(
        "{{ax_form_name}}", form.name
    ).replace(
        "{{ax_db_name}}", form.db_name
    ).replace(
        "{{guid}}", str(form.row_guid)
    )

    tags = re.findall("{{(.*?)}}", tom_label_modified)
    for field in form.fields:
        if field.db_name in tags:
            tom_label_modified = tom_label_modified.replace(
                "{{" + field.db_name + "}}", str(field.value)
            )

    return tom_label_modified


async def get_ax_host():
    """ Returns ax instance url with method and port. From app.yaml """
    return os.environ.get('AX_URL')


async def fetch(url):
    """ fetchs web-page by url """
    try:
        timeout = aiohttp.ClientTimeout(total=20)
        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=timeout) as response:
                return await response.text()
    except (asyncio.TimeoutError, ConnectionRefusedError):
        return None


async def post_json(url, data, timeout=20):
    """ Make post request with json data """
    try:
        timeout = aiohttp.ClientTimeout(total=timeout)
        async with aiohttp.ClientSession() as session:
            async with session.post(
                    url, json=data, timeout=timeout) as response:
                return await response.text()
    except (asyncio.TimeoutError, ConnectionRefusedError):
        return None



"""Defines Users Scheme and all mutations"""

import os
import sys
import re
import json
import zipfile
import shutil
import ast
import uuid
import asyncio
import urllib.request
from packaging import version
from loguru import logger
import graphene
import aiopubsub
import ruamel.yaml

from backend.model import AxForm, AxMetric, AxField, AxGrid, AxRole, AxState,\
    AxColumn, AxAction, AxAction2Role, AxState2Role, AxRoleFieldPermission,\
    Ax1tomReference, AxPage, AxFieldType
import backend.model as ax_model
import backend.dialects as ax_dialects
import backend.misc as ax_misc
from backend.auth import ax_admin_only
import backend.pubsub as ax_pubsub

this = sys.modules[__name__]

current_modal_guid = None
yaml = ruamel.yaml.YAML()
yaml.encoding = "utf-8"
yaml.allow_unicode = True
yaml.preserve_quotes = True


async def terminal_log(msg):
    """ Method used in AxTerminal. It sends message to vue via ws """
    if not this.current_modal_guid:
        return False

    ax_pubsub.publisher.publish(
        aiopubsub.Key('console_log'),
        {
            'text': msg,
            'modal_guid': current_modal_guid
        })
    await asyncio.sleep(0.01)


async def get_child_forms(db_session, form):
    """ returns flat list of all forms in tree """
    err = "marketplace_schema -> get_child_forms"
    with ax_model.try_catch(db_session, err) as db_session:
        ret_list = [form]
        sub_forms = db_session.query(AxForm).filter(
            AxForm.parent == form.guid
        ).all()

        for sub_form in sub_forms:
            # ret_list.append(sub_form)
            ret_list += await get_child_forms(db_session, sub_form)

        return ret_list


async def get_child_pages(db_session, page):
    """ returns flat list of all sub pages in tree """
    err = "marketplace_schema -> get_child_pages"
    with ax_model.try_catch(db_session, err) as db_session:
        ret_list = [page]
        sub_pages = db_session.query(AxPage).filter(
            AxPage.parent == page.guid
        ).all()

        for sub_page in sub_pages:
            # ret_list.append(sub_page)
            ret_list += await get_child_pages(db_session, sub_page)

        return ret_list


async def get_folders_and_forms(form_list):
    """ Splits flat form list to forms and folders """
    forms = []
    folders = []
    for form in form_list:
        if form.is_folder:
            folders.append(form)
        else:
            forms.append(form)
    return forms, folders


async def get_folders_dump(folders):
    """ returns list with serilized AxForm that are folders """
    ret_list = []
    for folder in folders:
        ret_list.append({
            "guid": str(folder.guid),
            "name": folder.name,
            "parent": str(folder.parent),
            "position": str(folder.position)
        })
    return ret_list


async def get_forms_db_names(forms):
    """ return flat list of form db_names """
    ret_list = []
    for form in forms:
        ret_list.append(form.db_name)
    return ret_list


async def save_page_py(page, pages_directory):
    """ Saves page code as file in package """
    if page.code:
        page_name = str(page.guid)
        if page.db_name:
            page_name = page.db_name
        page_md_path = pages_directory / f'page_{page_name}.md'
        with open(page_md_path, 'w', encoding="utf-8") as page_file:
            page_file.write(page.code)


async def dump_pages(db_session, package_directory, root_page):
    """ for each pages creates file and serilizes data, returns list """
    err = "marketplace_schema -> dump_pages"
    with ax_model.try_catch(db_session, err) as db_session:
        # Create AxPages folder
        ret_list = []
        pages_directory = package_directory / '_pages'
        os.makedirs(pages_directory)
        all_pages = await get_child_pages(db_session, root_page)

        for page in all_pages:
            ret_list.append({
                "guid": str(page.guid),
                "name": page.name,
                "db_name": page.db_name,
                "position": page.position,
                "parent": str(page.parent)
            })
            await save_page_py(page=page, pages_directory=pages_directory)

        return ret_list


async def dump_form(db_session, package_directory, form, include_data):
    """
        Creates dump of form and all related objects in separate folder ->
        /<db_name>/<db_name>.yaml contains most of data.
        # AxForm
        # AxField[]
        # AxMetric[]
        # AxGrid[], AxColumn[] - grid code saved as separate py files
        # AxRole[]
        # AxStates[], AxState2Role[]
        # AxActions[] - action code saved
        # AxRoleFieldPermission[]
        # Ax1tomReference[] - as tom_reference.yaml
        # SQL data - as data.yaml
    """
    err = "marketplace_schema -> dump_form"
    with ax_model.try_catch(db_session, err) as db_session:

        # Create directory
        form_path = package_directory / form.db_name
        if os.path.exists(form_path) is False:
            os.makedirs(form_path)
        form_yaml_path = form_path / f"{form.db_name}.yaml"
        with open(form_yaml_path, 'w', encoding="utf-8") as yaml_file:
            yaml_dict = {}
            # AxForm
            yaml_dict["AxForm"] = {
                "guid": str(form.guid),
                "name": form.name,
                "db_name": form.db_name,
                "parent": str(form.parent),
                "position": form.position,
                "tom_label": form.tom_label,
                "icon": form.icon,
            }

            # AxField[]
            yaml_dict["AxFields"] = []
            metric_fields = []
            for field in form.fields:
                yaml_dict["AxFields"].append({
                    "guid": str(field.guid),
                    "name": field.name,
                    "db_name": field.db_name,
                    "position": field.position,
                    "options_json": field.options_json,
                    "private_options_json": field.private_options_json,
                    "field_type_tag": field.field_type_tag,
                    "is_tab": field.is_tab,
                    "is_required": field.is_required,
                    "is_whole_row": field.is_whole_row,
                    "parent": str(field.parent)
                })

                if field.field_type_tag in ['AxNum']:
                    metric_fields.append(field)

            # AxMetric
            yaml_dict["AxMetrics"] = []
            for met_field in metric_fields:
                options = json.loads(met_field.private_options_json)
                key = options.get('counterKey')
                if key:
                    ax_metric = db_session.query(AxMetric).filter(
                        AxMetric.key == key
                    ).first()

                    if ax_metric:
                        yaml_dict["AxMetrics"].append({
                            "guid": str(ax_metric.guid),
                            "key": ax_metric.key,
                            "value": ax_metric.value
                        })

            # AxGrid[], AxColumns[]
            yaml_dict["AxGrids"] = []
            for grid in form.grids:
                columns = []
                for col in grid.columns:
                    columns.append({
                        "guid": str(col.guid),
                        "position": col.position,
                        "options_json": col.options_json,
                        "field_guid": str(col.field_guid),
                        "column_type": col.column_type,
                        "aggregation_type": col.aggregation_type
                    })

                yaml_dict["AxGrids"].append({
                    "guid": str(grid.guid),
                    "name": grid.name,
                    "db_name": grid.db_name,
                    "position": grid.position,
                    "options_json": grid.options_json,
                    "is_default_view": grid.is_default_view,
                    "columns": columns
                })

                if grid.code:
                    grid_py_path = form_path / f'grid_{grid.db_name}.py'
                    with open(grid_py_path, 'w', encoding="utf-8") as grid_file:
                        grid_file.write(grid.code)

            # AxRoles[]
            yaml_dict["AxRoles"] = []
            for role in form.roles:
                yaml_dict["AxRoles"].append({
                    "guid": str(role.guid),
                    "name": role.name,
                    "icon": role.icon,
                    "is_dynamic": role.is_dynamic
                })
                if role.code:
                    role_py_path = form_path / f'role_{role.db_name}.py'
                    with open(role_py_path, 'w', encoding="utf-8") as role_file:
                        role_file.write(role.code)

            # AxStates[]
            yaml_dict["AxStates"] = []
            for state in form.states:
                role_guids = []
                for state_role in state.roles:
                    role_guids.append(str(state_role.guid))
                yaml_dict["AxStates"].append({
                    "guid": str(state.guid),
                    "name": state.name,
                    "is_start": state.is_start,
                    "is_deleted": state.is_deleted,
                    "is_all": state.is_all,
                    "x": state.x,
                    "y": state.y,
                    "roles": role_guids
                })

            # AxActions[]
            yaml_dict["AxActions"] = []
            for action in form.actions:
                role_guids = []
                for action_role in action.roles:
                    role_guids.append(str(action_role.guid))
                yaml_dict["AxActions"].append({
                    "guid": str(action.guid),
                    "name": action.name,
                    "db_name": action.db_name,
                    "from_state_guid": str(action.from_state_guid),
                    "to_state_guid": str(action.to_state_guid),
                    "confirm_text": action.confirm_text,
                    "close_modal": action.close_modal,
                    "icon": action.icon,
                    "radius": action.radius,
                    "roles": role_guids
                })

                # save_action_py
                if action.code:
                    action_id = str(action.guid)
                    if action.db_name:
                        action_id = action.db_name
                    action_py_path = form_path / f'action_{action_id}.py'
                    with open(action_py_path, 'w', encoding="utf-8") as a_file:
                        a_file.write(action.code)

            # AxRoleFieldPermission[]
            yaml_dict["AxRoleFieldPermission"] = []
            for perm in form.permissions:
                yaml_dict["AxRoleFieldPermission"].append({
                    "role_guid": str(perm.role_guid),
                    "state_guid": str(perm.state_guid),
                    "field_guid": str(perm.field_guid),
                    "read": perm.read,
                    "edit": perm.edit
                })

            yaml.dump(yaml_dict, yaml_file)

            if include_data:
                # Ax1tomReference[]
                tom_references = db_session.query(Ax1tomReference).filter(
                    Ax1tomReference.form_guid == form.guid
                ).all()
                if tom_references:
                    clean_tom = []
                    for tom in tom_references:
                        clean_tom.append({
                            "guid": str(tom.guid),
                            "field_guid": str(tom.field_guid),
                            "row_guid": str(tom.row_guid),
                            "child_guid": str(tom.child_guid)
                        })
                    tom_path = form_path / f'tom_reference.yaml'
                    with open(tom_path, 'w', encoding="utf-8") as tom_file:
                        yaml.dump(clean_tom, tom_file)

                # SQL data
                data_result = db_session.execute(
                    f'SELECT * FROM "{form.db_name}"')
                if data_result:
                    def clear_value(val):
                        # if isinstance(val, uuid.UUID):
                        if val is None:
                            return None
                        if isinstance(val, (int, float, complex, bool)):
                            return val
                        return str(val)

                    clean_result = []
                    for row in data_result:
                        clean_row = {}
                        for key, value in row.items():
                            clean_row[key] = clear_value(value)
                        clean_result.append(clean_row)

                    data_path = form_path / f'data.yaml'
                    with open(data_path, 'w', encoding="utf-8") as data_file:
                        yaml.dump(clean_result, data_file)


async def do_create_readme(root_page, package_directory):
    """ Create Readme.md from root_page code """
    readme_path = package_directory / f'README.md'
    with open(readme_path, 'w', encoding="utf-8") as readme_file:
        readme_file.write(root_page.code)


class CreateMarketplaceApplication(graphene.Mutation):
    """ Creates archive with marketpalce application"""
    class Arguments:  # pylint: disable=missing-docstring
        folder_guid = graphene.String()
        name = graphene.String()
        version = graphene.String()
        db_name = graphene.String()
        root_page = graphene.String(required=False)
        include_data = graphene.Boolean()
        create_readme = graphene.Boolean()

    ok = graphene.Boolean()
    download_url = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "marketplace_schema -> CreateMarketplaceApplication"
        code_name = args.get('db_name')
        root_page_db_name = args.get('root_page')
        unique_guid = str(uuid.uuid4())
        include_data = args.get('include_data')
        create_readme = args.get('create_readme')

        with ax_model.try_catch(info.context['session'], err) as db_session:
            # Create tmp folder for new package
            archive_directory = ax_misc.tmp_root_dir / unique_guid
            package_directory = archive_directory / 'archive_me' / code_name
            dir_to_zip = archive_directory / 'archive_me'
            os.makedirs(package_directory)

            # Get AxForms and folders to dump
            root_folder = db_session.query(AxForm).filter(
                AxForm.guid == ax_misc.guid_or_none(
                    str(args.get('folder_guid')))
            ).first()
            if not root_folder:
                raise Exception('Cant find app root folder')

            # AxPage[]
            # Get AxPage if db_name is provided
            pages_data = []
            root_page = None
            root_page_guid = None
            if root_page_db_name:
                root_page = db_session.query(AxPage).filter(
                    AxPage.db_name == root_page_db_name
                ).first()

                if not root_page:
                    raise Exception('Cant find app root page')

                # Сreate readme.md
                if create_readme:
                    await do_create_readme(
                        root_page=root_page,
                        package_directory=package_directory)

                root_page_guid = str(root_page.guid)
                pages_data = await dump_pages(
                    db_session=db_session,
                    package_directory=package_directory,
                    root_page=root_page)

            all_forms = await get_child_forms(
                db_session=db_session, form=root_folder)
            app_forms, app_folders = await get_folders_and_forms(all_forms)

            # Create app.yaml
            app_yaml_path = package_directory / 'ax_app.yaml'
            with open(app_yaml_path, 'w', encoding="utf-8") as yaml_file:
                app_info = {
                    "Code name": code_name,
                    "Root folder": args.get('folder_guid'),
                    "Forms": await get_forms_db_names(app_forms),
                    "Folders": await get_folders_dump(app_folders),
                    "Root page": root_page_guid,
                    "Pages": pages_data,
                    "Ax version": os.environ.get('AX_VERSION')
                }
                yaml.dump(app_info, yaml_file)

            for form in app_forms:
                await dump_form(
                    db_session=db_session,
                    package_directory=package_directory,
                    form=form,
                    include_data=include_data)

            # version_str = str(app_version).replace(',', '-')
            # app_file_name = f'{code_name}-{version_str}.zip'
            app_file_name = f'{code_name}.zip'
            output_file_path = archive_directory / app_file_name

            await ax_misc.zip_folder(
                output_filename=output_file_path,
                source_dir=dir_to_zip)

            shutil.rmtree(dir_to_zip)

            url = f'/api/file/null/null/null/{unique_guid}/{app_file_name}'
            return CreateMarketplaceApplication(download_url=url, ok=True)



async def create_page(
        db_session, page, package_directory, ax_root_guid, app_root_guid):
    """ Create AxPage from application package data """
    err = "marketplace_schema -> create_page"
    with ax_model.try_catch(db_session, err) as db_session:
        code = None
        page_name = str(page['guid'])
        if page['db_name']:
            page_name = page['db_name']
        page_md_path = package_directory / '_pages' / f'page_{page_name}.md'
        if os.path.exists(page_md_path):
            with open(page_md_path, 'r', encoding="utf-8") as page_file:
                code = page_file.read()

        existing_page = db_session.query(AxPage).filter(
            AxPage.guid == ax_misc.guid_or_none(page['guid'])
        ).first()

        if existing_page:
            existing_page.name = page['name']
            existing_page.db_name = page['db_name']
            existing_page.position = int(page['position'] or 0)
            existing_page.parent = ax_misc.guid_or_none(page['parent'])
            existing_page.code = code
        else:
            cur_parent = page['parent']
            if page['guid'] == app_root_guid:
                cur_parent = ax_root_guid

            new_page = AxPage()
            new_page.guid = ax_misc.guid_or_none(page['guid'])
            new_page.name = page['name']
            new_page.db_name = page['db_name']
            new_page.position = int(page['position'] or 0)
            new_page.parent = ax_misc.guid_or_none(cur_parent)
            new_page.code = code
            db_session.add(new_page)


async def create_folder(db_session, folder):
    """ Create or update AxForm that is folder """
    err = "marketplace_schema -> create_folder"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_folder = db_session.query(AxForm).filter(
            AxForm.guid == ax_misc.guid_or_none(folder['guid'])
        ).filter(
            AxForm.is_folder.is_(True)
        ).first()

        if existing_folder:
            existing_folder.name = folder['name']
        else:
            new_folder = AxForm()
            new_folder.guid = ax_misc.guid_or_none(folder['guid'])
            new_folder.name = folder['name']
            new_folder.is_folder = True
            new_folder.position = int(folder['position'] or 0)
            new_folder.parent = ax_misc.guid_or_none(folder['parent'])
            db_session.add(new_folder)


async def create_form(db_session, form):
    """ Create or update AxForm """
    err = "marketplace_schema -> create_form"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_form = db_session.query(AxForm).filter(
            AxForm.guid == ax_misc.guid_or_none(form['guid'])
        ).first()
        new_form = None
        form_is_new = False

        if existing_form:
            existing_form.name = form['name']
            existing_form.tom_label = form['tom_label']
            existing_form.icon = form['icon']
        else:
            form_is_new = True
            new_form = AxForm()
            new_form.guid = ax_misc.guid_or_none(form['guid'])
            new_form.name = form['name']
            new_form.db_name = form['db_name']
            new_form.position = int(form['position'] or 0)
            new_form.parent = ax_misc.guid_or_none(form['parent'])
            new_form.tom_label = form['tom_label']
            new_form.icon = form['icon']
            db_session.add(new_form)

            await ax_dialects.dialect.create_data_table(
                db_session=db_session,
                db_name=form['db_name'])

            db_session.commit()

        ret_form = existing_form or new_form
        return form_is_new, ret_form


async def create_metric(db_session, metric):
    """ Create AxMetric if not exists """
    err = "marketplace_schema -> create_metric"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_metric = db_session.query(AxMetric).filter(
            AxMetric.guid == ax_misc.guid_or_none(metric['guid'])
        ).first()

        if not existing_metric:
            new_metric = AxMetric()
            new_metric.guid = ax_misc.guid_or_none(metric['guid'])
            new_metric.key = metric['key']
            new_metric.value = metric['value']
            db_session.add(new_metric)


async def create_field(db_session, ax_form, field):
    """ Create or update AxField """
    err = "marketplace_schema -> create_field"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_field = db_session.query(AxField).filter(
            AxField.guid == ax_misc.guid_or_none(field['guid'])
        ).first()

        if existing_field:
            existing_field.name = field['name']
            existing_field.options_json = field['options_json']
            existing_field.private_options_json = field['private_options_json']
            existing_field.is_required = field['is_required']
            existing_field.is_whole_row = field['is_whole_row']
            existing_field.parent = ax_misc.guid_or_none(field['parent'])
            existing_field.position = int(field['position'] or 0)
        else:
            new_field = AxField()
            new_field.guid = ax_misc.guid_or_none(field['guid'])
            new_field.form_guid = ax_form.guid
            new_field.name = field['name']
            new_field.db_name = field['db_name']
            new_field.options_json = field['options_json']
            new_field.private_options_json = field['private_options_json']
            new_field.field_type_tag = field['field_type_tag']
            new_field.is_tab = field['is_tab']
            new_field.is_required = field['is_required']
            new_field.is_whole_row = field['is_whole_row']
            new_field.parent = ax_misc.guid_or_none(field['parent'])
            new_field.position = int(field['position'] or 0)
            db_session.add(new_field)

            # Field is not tab
            if field['field_type_tag']:
                ax_field_type = db_session.query(AxFieldType).filter(
                    AxFieldType.tag == field['field_type_tag']
                ).first()

                if not ax_field_type.is_virtual:
                    await ax_dialects.dialect.add_column(
                        db_session=db_session,
                        table=ax_form.db_name,
                        db_name=field['db_name'],
                        type_name=ax_field_type.value_type)

                    db_session.commit()


async def create_column(db_session, grid_guid, column):
    """ Create or update AxColumn """
    err = "marketplace_schema -> create_column"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_column = db_session.query(AxColumn).filter(
            AxColumn.guid == ax_misc.guid_or_none(column['guid'])
        ).first()
        new_column = None

        if existing_column:
            existing_column.position = int(column['position'] or 0)
            existing_column.options_json = column['options_json']
            existing_column.column_type = column['column_type']
            existing_column.aggregation_type = column['aggregation_type']
        else:
            new_column = AxColumn()
            new_column.guid = ax_misc.guid_or_none(column['guid'])
            new_column.position = int(column['position'] or 0)
            new_column.options_json = column['options_json']
            new_column.field_guid = ax_misc.guid_or_none(column['field_guid'])
            new_column.grid_guid = ax_misc.guid_or_none(grid_guid)
            new_column.column_type = column['column_type']
            new_column.aggregation_type = column['aggregation_type']
            db_session.add(new_column)


async def create_grid(db_session, form_directory, ax_form, grid):
    """ Create or update AxGrid """
    err = "marketplace_schema -> create_grid"
    with ax_model.try_catch(db_session, err) as db_session:
        code = None
        existing_grid = db_session.query(AxGrid).filter(
            AxGrid.guid == ax_misc.guid_or_none(grid['guid'])
        ).first()
        new_grid = None

        grid_py_path = form_directory / f'grid_{grid["db_name"]}.py'
        if os.path.exists(grid_py_path):
            with open(grid_py_path, 'r', encoding="utf-8") as grid_file:
                code = grid_file.read()

        if existing_grid:
            existing_grid.name = grid['name']
            existing_grid.db_name = grid['db_name']
            existing_grid.position = int(grid['position'] or 0)
            existing_grid.options_json = grid['options_json']
            existing_grid.code = code
            existing_grid.is_default_view = grid['is_default_view']
        else:
            new_grid = AxGrid()
            new_grid.guid = ax_misc.guid_or_none(grid['guid'])
            new_grid.name = grid['name']
            new_grid.db_name = grid['db_name']
            new_grid.position = int(grid['position'] or 0)
            new_grid.options_json = grid['options_json']
            new_grid.code = code
            new_grid.form_guid = ax_form.guid
            new_grid.is_default_view = grid['is_default_view']
            db_session.add(new_grid)

    for column in grid['columns']:
        await create_column(
            db_session=db_session,
            grid_guid=grid['guid'],
            column=column)

    return existing_grid or new_grid


async def create_role(db_session, ax_form, role, form_path):
    """ Create or update AxRole """
    err = "marketplace_schema -> create_role"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_role = db_session.query(AxRole).filter(
            AxRole.guid == ax_misc.guid_or_none(role['guid'])
        ).first()
        new_role = None

        code = None
        role_db_name = re.sub(r'[\W_]+', '', role['name'])
        role_py_path = form_path / f'role_{role_db_name}.py'
        if os.path.exists(role_py_path):
            with open(role_py_path, 'r', encoding="utf-8") as role_file:
                code = role_file.read()

        if existing_role:
            existing_role.name = role['name']
            existing_role.icon = role.get('icon', None)
        else:
            new_role = AxRole()
            new_role.guid = ax_misc.guid_or_none(role['guid'])
            new_role.name = role['name']
            new_role.form_guid = ax_form.guid
            new_role.icon = role.get('icon', None)
            new_role.is_dynamic = role.get('is_dynamic', False)
            new_role.code = code
            db_session.add(new_role)


async def create_state2role(db_session, state_guid, role_guid):
    """ Create or update AxState2Role """
    err = "marketplace_schema -> create_state2role"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_s2r = db_session.query(AxState2Role).filter(
            AxState2Role.state_guid == ax_misc.guid_or_none(state_guid)
        ).filter(
            AxState2Role.role_guid == ax_misc.guid_or_none(role_guid)
        ).first()

        if not existing_s2r:
            new_s2r = AxState2Role()
            new_s2r.state_guid = ax_misc.guid_or_none(state_guid)
            new_s2r.role_guid = ax_misc.guid_or_none(role_guid)
            db_session.add(new_s2r)


async def create_state(db_session, ax_form, state):
    """ Create or update AxState """
    err = "marketplace_schema -> create_state"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_state = db_session.query(AxState).filter(
            AxState.guid == ax_misc.guid_or_none(state['guid'])
        ).first()
        new_state = None

        if existing_state:
            existing_state.name = state['name']
            existing_state.x = float(state['x'] or 0)
            existing_state.y = float(state['y'] or 0)
        else:
            new_state = AxState()
            new_state.guid = ax_misc.guid_or_none(state['guid'])
            new_state.name = state['name']
            new_state.form_guid = ax_form.guid
            new_state.is_start = state['is_start']
            new_state.is_deleted = state['is_deleted']
            new_state.is_all = state['is_all']
            new_state.x = float(state['x'] or 0)
            new_state.y = float(state['y'] or 0)
            db_session.add(new_state)

        for role_guid in state['roles']:
            await create_state2role(
                db_session=db_session,
                state_guid=state['guid'],
                role_guid=role_guid)


async def create_action2role(db_session, action_guid, role_guid):
    """ Create or update AxAction2Role """
    err = "marketplace_schema -> create_action2role"
    with ax_model.try_catch(db_session, err) as db_session:
        existing_a2r = db_session.query(AxAction2Role).filter(
            AxAction2Role.action_guid == ax_misc.guid_or_none(action_guid)
        ).filter(
            AxAction2Role.role_guid == ax_misc.guid_or_none(role_guid)
        ).first()

        if not existing_a2r:
            new_a2r = AxAction2Role()
            new_a2r.action_guid = ax_misc.guid_or_none(action_guid)
            new_a2r.role_guid = ax_misc.guid_or_none(role_guid)
            db_session.add(new_a2r)


async def drop_actions(db_session, ax_form):
    """ Deletes all AxActions and AxAction2Role of form """
    err = "marketplace_schema -> drop_actions"
    with ax_model.try_catch(db_session, err) as db_session:
        for action in ax_form.actions:
            db_session.query(AxAction2Role).filter(
                AxAction2Role.action_guid == action.guid).delete()
            db_session.delete(action)


async def create_action(db_session, form_directory, ax_form, action):
    """ Creates new AxAction """
    err = "marketplace_schema -> create_action"
    with ax_model.try_catch(db_session, err) as db_session:
        code = None
        action_id = str(action['guid'])
        if action['db_name']:
            action_id = action['db_name']
        action_py_path = form_directory / f'action_{action_id}.py'

        if os.path.exists(action_py_path):
            with open(action_py_path, 'r', encoding="utf-8") as action_file:
                code = action_file.read()

        new_action = AxAction()
        new_action.guid = ax_misc.guid_or_none(action['guid'])
        new_action.name = action['name']
        new_action.db_name = action['db_name']
        new_action.form_guid = ax_form.guid
        new_action.from_state_guid = ax_misc.guid_or_none(
            action['from_state_guid'])
        new_action.to_state_guid = ax_misc.guid_or_none(
            action['to_state_guid'])
        new_action.code = code
        new_action.confirm_text = action['confirm_text']
        new_action.close_modal = action['close_modal']
        new_action.icon = action['icon']
        new_action.radius = float(action['radius'] or 0)
        db_session.add(new_action)

        for role_guid in action['roles']:
            await create_action2role(
                db_session=db_session,
                action_guid=action['guid'],
                role_guid=role_guid)


async def create_role_field_perm(db_session, ax_form, perm):
    """ AxRoleFieldPermission creaes new """
    err = "marketplace_schema -> create_role_field_perm"
    with ax_model.try_catch(db_session, err) as db_session:
        new_perm = AxRoleFieldPermission()
        new_perm.role_guid = ax_misc.guid_or_none(str(perm['role_guid']))
        new_perm.state_guid = ax_misc.guid_or_none(str(perm['state_guid']))
        new_perm.field_guid = ax_misc.guid_or_none(str(perm['field_guid']))
        new_perm.form_guid = ax_form.guid
        new_perm.read = perm['read']
        new_perm.edit = perm['edit']
        db_session.add(new_perm)


async def create_tom_references(db_session, ax_form, form_directory):
    """ # Ax1tomReference[] - only insert if not exists """
    err = "marketplace_schema -> create_tom_references"
    with ax_model.try_catch(db_session, err) as db_session:
        tom_yaml_path = form_directory / 'tom_reference.yaml'
        if not os.path.exists(tom_yaml_path):
            return False

        with open(tom_yaml_path, 'r', encoding="utf-8") as tom_yaml_file:
            tom_yaml = yaml.load(tom_yaml_file)
            for ref in tom_yaml:
                new_ref = Ax1tomReference()
                new_ref.form_guid = ax_form.guid
                new_ref.field_guid = ax_misc.guid_or_none(ref['field_guid'])
                new_ref.row_guid = ax_misc.guid_or_none(ref['row_guid'])
                new_ref.child_guid = ax_misc.guid_or_none(ref['child_guid'])
                new_ref.field_guid = ax_misc.guid_or_none(ref['field_guid'])
                db_session.add(new_ref)


def clean_yaml_value(row, field):
    """ When json is dumped to yaml is forced with double qoutes,
        but when readed - qoutes stays in string. We must remove them """
    value = row[field.db_name]
    if value and field.field_type.value_type == 'JSON':
        try:
            value = json.loads(value)
        except:
            try:
                value = ast.literal_eval(value)
            except:
                value_str = str(value)
                logger.error(
                    f'clean_yaml_value -> could not decode json {value_str}')
                value = None
    return value


async def insert_data(db_session, ax_form, form_directory):
    """ Inserts data from package """
    err = "marketplace_schema -> insert_data"
    with ax_model.try_catch(db_session, err) as db_session:
        data_yaml_path = form_directory / 'data.yaml'
        if not os.path.exists(data_yaml_path):
            return False

        with open(data_yaml_path, 'r', encoding="utf-8") as data_yaml_file:
            data_yaml = yaml.load(data_yaml_file)
            for row in data_yaml:
                tobe_form = db_session.query(AxForm).filter(
                    AxForm.guid == ax_form.guid
                ).first()

                for field in tobe_form.db_fields:
                    if field.db_name in row.keys():
                        field.value = clean_yaml_value(row=row, field=field)
                        field.needs_sql_update = True

                await ax_dialects.dialect.insert(
                    db_session=db_session,
                    form=tobe_form,
                    to_state_name=row["axState"],
                    new_guid=ax_misc.guid_or_none(row['guid']))


async def create_form_objects(db_session, form_name, package_directory):
    """ Creates or updates
        # AxForm - when update position and parent is ignored
        # AxField[] - when update - db_name is ignored
        # AxMetric[] - only insert if does not exist
        # AxGrid[], AxColumn[]
        # AxRole[]
        # AxStates[], AxState2Role[]
        # AxActions[] - deletes, creates new
        # AxRoleFieldPermission[] - deletes, creaes new
        # Ax1tomReference[] - insert if first install
        # SQL data - insert if first install
        """
    err = "marketplace_schema -> create_form_objects"
    with ax_model.try_catch(db_session, err) as db_session:
        form_directory = package_directory / form_name
        yaml_path = form_directory / f'{form_name}.yaml'
        with open(yaml_path, 'r', encoding="utf-8") as form_yaml_file:
            form_yaml = yaml.load(form_yaml_file)

            form_db_name = form_yaml["AxForm"]["db_name"]
            await terminal_log(f'\nCreating form [{form_db_name}] \n')

            # AxForm
            form_is_new = False
            form_is_new, ax_form = await create_form(
                db_session=db_session, form=form_yaml["AxForm"])

            # AxField[]
            await terminal_log(f'AxFields:\n')
            for field in form_yaml["AxFields"]:
                await create_field(
                    db_session=db_session,
                    ax_form=ax_form,
                    field=field)
                await terminal_log(f'    {field["db_name"]} ✔\n')

            # AxMetric[]
            for metric in form_yaml["AxMetrics"]:
                await create_metric(db_session=db_session, metric=metric)

            # AxGrid[], AxColumn[]
            await terminal_log(f'AxGrids:\n')
            for grid in form_yaml["AxGrids"]:
                await create_grid(
                    db_session=db_session,
                    form_directory=form_directory,
                    ax_form=ax_form,
                    grid=grid)
                await terminal_log(f'    {grid["db_name"]} ✔\n')

            # AxRole[]
            await terminal_log(f'AxRoles:\n')
            for role in form_yaml['AxRoles']:
                await create_role(
                    db_session=db_session,
                    ax_form=ax_form,
                    role=role,
                    form_path=form_directory)
                await terminal_log(f'    {role["name"]} ✔\n')

            # AxStates[]
            await terminal_log(f'AxStates:\n')
            for state in form_yaml['AxStates']:
                await create_state(
                    db_session=db_session,
                    ax_form=ax_form,
                    state=state)
                await terminal_log(f'    {state["name"]} ✔\n')

            # AxActions[] - deletes, creates new
            await terminal_log(f'AxActions:\n')
            await drop_actions(db_session=db_session, ax_form=ax_form)
            for action in form_yaml['AxActions']:
                await create_action(
                    db_session=db_session,
                    form_directory=form_directory,
                    ax_form=ax_form,
                    action=action)
                await terminal_log(f'    {action["name"]} ✔\n')

            db_session.commit()

            # AxRoleFieldPermission[] - deletes, creaes new
            for ax_perm in ax_form.permissions:
                db_session.delete(ax_perm)
            for perm in form_yaml['AxRoleFieldPermission']:
                await create_role_field_perm(
                    db_session,
                    ax_form,
                    perm)

            if form_is_new:
                await terminal_log(f'Creating data')
                # Ax1tomReference[] - insert if first install
                await create_tom_references(
                    db_session=db_session,
                    ax_form=ax_form,
                    form_directory=form_directory)

                # SQL data - insert if first install
                await insert_data(
                    db_session=db_session,
                    ax_form=ax_form,
                    form_directory=form_directory)
                await terminal_log(f' ->  ✔\n')
            else:
                await terminal_log(
                    f'Form already exists, data was not inserted')



async def check_existing_form(db_session, package_directory, form_db_name):
    """ Checks if AxForm with such db_name already exists. Aborts installation
        if found """
    form_directory = package_directory / form_db_name
    yaml_path = form_directory / f'{form_db_name}.yaml'
    form_guid = None
    with open(yaml_path, 'r', encoding="utf-8") as form_yaml_file:
        form_yaml = yaml.load(form_yaml_file)
        form_guid = form_yaml["AxForm"]["guid"]

    existing_form = db_session.query(AxForm).filter(
        AxForm.db_name == form_db_name
    ).first()

    if existing_form.guid != ax_misc.guid_or_none(form_guid):
        err = (f'Can not create AxForm [{form_db_name}]. Form with same '
               f'db_name already exists.\nInstallation is aborted ☠️')
        logger.exception(err)
        await terminal_log(f'\n\n\n{err}')
        raise Exception(err)


async def check_package(db_session, app_yaml, package_directory):
    """ Check if package can be installed """
    # Check ax version
    code_name = app_yaml["Code name"]
    app_ax_version = app_yaml["Ax version"]
    system_ax_version = os.environ.get('AX_VERSION')
    if version.parse(system_ax_version) < version.parse(app_ax_version):
        msg = (f'Error installing application [{code_name}].'
               f'Current version of Ax [{system_ax_version}] must be upgraded '
               f'to [{app_ax_version}] for this application. '
               f'\nInstallation is aborted ☠️')
        logger.exception(msg)
        await terminal_log(f'\n\n\n{msg}')
        raise Exception(msg)

    # Check if table names of application is already present in database
    ax_forms = db_session.query(AxForm).with_entities(AxForm.db_name).all()
    existing_db_names = [form.db_name for form in ax_forms]
    for form_db_name in app_yaml['Forms']:
        if form_db_name in existing_db_names:
            # Check if guid of existing form is same as application form
            await check_existing_form(
                db_session=db_session,
                package_directory=package_directory,
                form_db_name=form_db_name)
        else:
            # Check if table exists
            if ax_model.engine.dialect.has_table(ax_model.engine, form_db_name):
                err = (f'Database already contains table [{form_db_name}],'
                       f'but it is not registered as AxForm'
                       f'.\nInstallation is aborted ☠️')
                logger.exception(err)
                await terminal_log(f'\n\n\n{err}')
                raise Exception(err)


async def install_from_dir(db_session, app_dir):
    """ Installs app from yaml files in /tmp folder """
    import backend.schema as ax_schema
    with open(app_dir / 'ax_app.yaml', 'r', encoding="utf-8") as app_yaml_file:
        app_yaml = yaml.load(app_yaml_file)

        # Check if package can be installed
        await check_package(
            db_session=db_session,
            app_yaml=app_yaml,
            package_directory=app_dir)

        # Create pages
        await terminal_log('Creating pages:\n')
        ax_root_page = db_session.query(AxPage).filter(
            AxPage.parent.is_(None)
        ).first()
        ax_root_guid = None
        if ax_root_page:
            ax_root_guid = str(ax_root_page.guid)

        app_root_guid = app_yaml.get('Root page', None)
        for page in app_yaml.get('Pages', None):
            await create_page(
                db_session=db_session,
                page=page,
                package_directory=app_dir,
                ax_root_guid=ax_root_guid,
                app_root_guid=app_root_guid)
            await terminal_log(f'    {page["name"]} ✔\n')

        # Create folders
        await terminal_log('Creating folders:\n')
        for folder in app_yaml['Folders']:
            await create_folder(db_session=db_session, folder=folder)
            await terminal_log(f'    {folder["name"]} ✔\n')

        # For each form - create form
        for form_name in app_yaml['Forms']:
            await create_form_objects(
                db_session=db_session,
                form_name=form_name,
                package_directory=app_dir)

        ax_schema.init_schema(db_session)
        await terminal_log('------------------\n Application is installed!')



async def install_from_repo(db_session, repo_name):
    """ Installs marketplace application from github repo """
    tmp_dir = ax_misc.tmp_root_dir / str(uuid.uuid4())
    os.makedirs(tmp_dir)
    tmp_zip = tmp_dir / "installed_app.zip"
    archive_dir = tmp_dir / "package_archive"

    # Read archive data to ram
    await terminal_log(f'Downloading master from repo [{repo_name}]!\n\n')
    url = f'https://github.com/{repo_name}/archive/master.zip'
    filedata = urllib.request.urlopen(url)
    data_to_write = filedata.read()

    # Write archvive data to tmp file
    with open(tmp_zip, 'wb') as file:
        file.write(data_to_write)

    # Unzip tmp file
    await terminal_log(f'Download success\n\n')
    with zipfile.ZipFile(tmp_zip, "r") as zip_ref:
        zip_ref.extractall(archive_dir)

    app_code_name = os.listdir(archive_dir)[0]
    app_dir = archive_dir / app_code_name
    await install_from_dir(db_session=db_session, app_dir=app_dir)



class InstallFromPackage(graphene.Mutation):
    """ Installs marketplace application from zip archive uploaded
        to tmp folder """
    class Arguments:  # pylint: disable=missing-docstring
        file_guid = graphene.String()
        modal_guid = graphene.String()

    ok = graphene.Boolean()
    msg = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - InstallFromPackage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            del info
            this.current_modal_guid = args.get('modal_guid')
            file_guid = args.get('file_guid')
            tmp_dir = ax_misc.tmp_root_dir / file_guid
            file_name = os.listdir(tmp_dir)[0]
            temp_path = tmp_dir / file_name
            archive_dir = tmp_dir / "package_archive"

            with zipfile.ZipFile(temp_path, "r") as zip_ref:
                zip_ref.extractall(archive_dir)

            app_code_name = os.listdir(archive_dir)[0]
            app_dir = archive_dir / app_code_name
            await install_from_dir(db_session=db_session, app_dir=app_dir)
            return InstallFromPackage(msg='InstallFromPackage - OK', ok=True)


class InstallFromRepo(graphene.Mutation):
    """ Installs marketplace application from github repo """
    class Arguments:  # pylint: disable=missing-docstring
        repo_name = graphene.String()
        modal_guid = graphene.String()

    ok = graphene.Boolean()
    msg = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring

        err = 'Error in gql mutation - InstallFromRepo.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            del info
            this.current_modal_guid = args.get('modal_guid')
            repo_name = args.get('repo_name')
            await install_from_repo(db_session=db_session, repo_name=repo_name)
            return InstallFromRepo(msg='InstallFromRepo - OK', ok=True)



class MarketplaceQuery(graphene.ObjectType):
    """ queryes"""


class MarketplaceMutations(graphene.ObjectType):
    """Contains all AxUser mutations"""
    create_marketplace_application = CreateMarketplaceApplication.Field()
    install_from_package = InstallFromPackage.Field()
    install_from_repo = InstallFromRepo.Field()



"""
AxNum field type functions - before, after / insert, update, delete

How to use in code:
    from datetime import datetime, timedelta
    expire = datetime.utcnow() + timedelta(days=366)
    ax.row.expire = expire.timestamp()

"""
import json


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """
    Executes python code from AxField.private_options

    Returns:
        Object: Returns updated value of current field"""
    del before_form, action, db_session, tobe_form, current_user

    options = json.loads(field.options_json)
    default_now = options.get('defaultNow', None)
    if default_now:
        field.force_sql_update = True

    return field.value



"""This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Defines AxPages Scheme and all mutations."""

import uuid
import graphene
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
import markdown2

from backend.misc import convert_column_to_string
from backend.model import GUID, AxPage, AxPage2Users, AxUser, AxGroup2Users
from backend.auth import ax_admin_only
from backend.schemas.types import Page, PositionInput

import backend.model as ax_model
# import backend.cache as ax_cache
# import backend.dialects as ax_dialects

convert_sqlalchemy_type.register(GUID)(convert_column_to_string)


def apply_markdown(code):
    """ apply markdown to code """
    return markdown2.markdown(code)


def get_allowed_pages_guids(db_session, user_guid):
    """ Get all users and users groups, then get all pages guids that
        user can view """
    page_guids = []
    user_and_groups = []

    # Add everyone group
    everyone_group = db_session.query(AxUser).filter(
        AxUser.is_everyone.is_(True)
    ).first()
    if everyone_group:
        user_and_groups.append(everyone_group.guid)

    if user_guid:
        # Add current user
        user_and_groups.append(uuid.UUID(user_guid))

        # Add all-users group
        all_users_group = db_session.query(AxUser).filter(
            AxUser.is_all_users.is_(True)
        ).first()
        user_and_groups.append(all_users_group.guid)

        # Get user groups
        group2user_result = db_session.query(AxGroup2Users).filter(
            AxGroup2Users.user_guid == uuid.UUID(user_guid)
        ).all()
        for g2u in group2user_result:
            if g2u.group_guid not in user_and_groups:
                user_and_groups.append(g2u.group_guid)

    page2users_result = db_session.query(AxPage2Users).filter(
        AxPage2Users.user_guid.in_(user_and_groups)
    ).all()
    for p2u in page2users_result:
        if p2u.page_guid not in page_guids:
            page_guids.append(p2u.page_guid)

    return page_guids


class CreatePage(graphene.Mutation):
    """ Creates AxPage """
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()
        parent = graphene.String()

    ok = graphene.Boolean()
    page = graphene.Field(Page)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - pages_schema -> CreatePage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            new_page = AxPage()
            new_page.name = args.get('name')
            new_page.parent = uuid.UUID(args.get('parent'))
            new_page.code = f"""<h1>{new_page.name}</h1>
Enter page content here"""
            db_session.add(new_page)
            db_session.flush()
            return CreatePage(page=new_page, ok=True)


class UpdatePage(graphene.Mutation):
    """ Update AxPage """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String(required=False, default_value=None)
        db_name = graphene.String(required=False, default_value=None)
        code = graphene.String(required=False, default_value=None)
        parent = graphene.String(required=False, default_value=None)

    ok = graphene.Boolean()
    page = graphene.Field(Page)
    msg = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        # avatar_tmp = args.get('avatar_tmp')
        guid = args.get('guid')
        name = args.get('name')
        db_name = args.get('db_name')
        code = args.get('code')
        parent = args.get('parent')

        err = 'Error in gql mutation - pages_schema -> UpdatePage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:

            ax_page = db_session.query(AxPage).filter(
                AxPage.guid == uuid.UUID(guid)
            ).first()

            if name:
                ax_page.name = name

            if db_name:
                already_exist_page = db_session.query(AxPage).filter(
                    AxPage.db_name == db_name
                ).first()
                if already_exist_page:
                    return UpdatePage(
                        page=None, ok=False, msg='pages.db-name-exists')
                ax_page.db_name = db_name

            if code:
                ax_page.code = code

            if parent:
                ax_page.parent = uuid.UUID(parent)

            db_session.commit()
            ax_page.html = apply_markdown(ax_page.code)
            return UpdatePage(page=ax_page, ok=True, msg=None)


class DeletePage(graphene.Mutation):
    """ Deletes AxPage """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - pages_schema -> DeletePage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_page = db_session.query(AxPage).filter(
                AxPage.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_page)
            return DeletePage(deleted=guid, ok=True)


class AddUserToPage(graphene.Mutation):
    """ Creates AxPage2Users """
    class Arguments:  # pylint: disable=missing-docstring
        user_guid = graphene.String()
        page_guid = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - pages_schema -> AddUserToPage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            user_guid = uuid.UUID(args.get('user_guid'))
            page_guid = uuid.UUID(args.get('page_guid'))

            page2user = db_session.query(AxPage2Users).filter(
                AxPage2Users.user_guid == user_guid
            ).filter(
                AxPage2Users.page_guid == page_guid
            ).first()
            if page2user:
                return AddUserToPage(ok=True)

            page2user = AxPage2Users()
            page2user.page_guid = page_guid
            page2user.user_guid = user_guid
            db_session.add(page2user)
            db_session.flush()

            return AddUserToPage(ok=True)


class RemoveUserFromPage(graphene.Mutation):
    """ Deletes AxPage2Users """
    class Arguments:  # pylint: disable=missing-docstring
        user_guid = graphene.String()
        page_guid = graphene.String()

    ok = graphene.Boolean()

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - pages_schema -> RemoveUserFromPage.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            user_guid = uuid.UUID(args.get('user_guid'))
            page_guid = uuid.UUID(args.get('page_guid'))

            db_session.query(AxPage2Users).filter(
                AxPage2Users.user_guid == user_guid
            ).filter(
                AxPage2Users.page_guid == page_guid
            ).delete()
            db_session.flush()

            return RemoveUserFromPage(ok=True)


class ChangePagesPositions(graphene.Mutation):
    """Change position and parent of multiple pages"""
    class Arguments:  # pylint: disable=missing-docstring
        positions = graphene.List(PositionInput)

    ok = graphene.Boolean()
    pages = graphene.List(Page)

    @ax_admin_only
    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "page_schema -> ChangePagesPositions"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            positions = args.get('positions')
            for position in positions:
                db_page = db_session.query(AxPage).filter(
                    AxPage.guid == uuid.UUID(position.guid)
                ).one()
                current_parent = None
                if position.parent != '#':
                    current_parent = uuid.UUID(position.parent)
                db_page.parent = current_parent
                db_page.position = position.position

            db_session.flush()

            query = Page.get_query(info)  # SQLAlchemy query
            pages_list = query.all()
            return ChangePagesPositions(pages=pages_list, ok=True)


class PagesSubscription(graphene.ObjectType):
    """ Pages ws subscriptions """


class PagesQuery(graphene.ObjectType):
    """AxPage queryes"""
    all_pages = graphene.List(
        Page,
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    page_data = graphene.Field(
        Page,
        guid=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    async def resolve_all_pages(self, info, update_time):
        """Get all pages"""
        del update_time
        current_user = info.context['user']
        err = 'Error in GQL query - resolve_all_pages.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:

            user_guid = current_user.get(
                "user_id", None) if current_user else None
            user_is_admin = current_user.get(
                'is_admin', False) if current_user else False

            query = Page.get_query(info)  # SQLAlchemy query
            if user_is_admin:
                page_list = query.all()
                return page_list
            else:
                pages_guids = get_allowed_pages_guids(
                    db_session=db_session, user_guid=user_guid)
                page_list = query.filter(AxPage.guid.in_(pages_guids)).all()
                return page_list

    async def resolve_page_data(self, info, guid=None, update_time=None):
        """Get specific page"""
        del update_time
        current_user = info.context['user']
        user_is_admin = current_user.get(
            'is_admin', False) if current_user else False
        err = 'Error in GQL query - resolve_page_data.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            user_guid = current_user.get(
                "user_id", None) if current_user else None
            query = Page.get_query(info)  # SQLAlchemy query
            page = None
            if guid:
                page = query.filter(AxPage.guid == uuid.UUID(guid)).first()
            else:
                page = query.filter(AxPage.parent.is_(None)).first()

            pages_guids = get_allowed_pages_guids(
                db_session=db_session, user_guid=user_guid)

            if page.guid in pages_guids or user_is_admin:
                page.html = apply_markdown(page.code)
                return page
            return None


class PagesMutations(graphene.ObjectType):
    """Contains all AxUser mutations"""
    create_page = CreatePage.Field()
    update_page = UpdatePage.Field()
    delete_page = DeletePage.Field()
    add_user_to_page = AddUserToPage.Field()
    remove_user_from_page = RemoveUserFromPage.Field()
    change_pages_positions = ChangePagesPositions.Field()
